# Horn-Schunck optical flow estimator

See: http://en.wikipedia.org/wiki/Horn%E2%80%93Schunck_method .  The abbreviation "hsof" stands for Horn-Schunck optical flow.

## Implementations

  * hsof-opt: C++ version (main source file: hsof-mic-opt.cpp).
  * hsof.py: Python version that uses numpy.
  * See the examples directory in the ParallelAccelerator package for the `@acc` and plain Julia versions.

## Notes

The benchmark reads two images of the same size; each image pixel is a brightness value between 0 and 255.
The optical flow between the two images is computed using the Horn-Schunk method and single-precision floating
point.  The resulting 2D array of 2D vectors is written to the file out.flo.  The input image filenames are given
as the first two arguments.

The algorithm is based on the following ideas:

First, it does the computation at multiple scales.  It starts by downsampling the images to the smallest scale,
in the code this is width 50 and same aspect ratio.  It computes the flow between the images at this scale using
the method below.  Then it goes to a slightly higher size, first interpolating the flow computed in the first step
from that size to the new size.  Second, it warps the second image back towards the first one using the current
flow.  Third, it uses the below method to compute the flow from the first image to the warped image, this results
in a correction to the flow, which is added to it.  Then the algorithm repeats for a larger size, and so on to
until it finally considers the original size of the images.  The finally flow resulting from that is the output.
The code below uses a fixed number of scales (10) where the ratio of the widths of subsequent iterations is
constant over all the iterations.

Second, to compute the flow at each step, the following ideas are used.  Brightness constancy is assumed, that is,
the total derivative of I is assumed to be zero.  Using the chain rule, this leads to the equation:

    (pd of I wrt x).(dx/dt)+(pd of I wrt y).(dy/dt)+(pd of I wrt t) = 0

Call the pd of I wrt x, Ix; and similarly for Iy and It.  Call dx/dt u and dy/dt v - these are the desired flow
values.  Thus we get:

    Ix.u+Iy.v+It = 0
     
We can consider this to be an equation to be applied to each pixel of the images, but it is one equation in two
unknowns, and thus does not solve the problem.  We must add further constraints.  The Horn-Schunck adds the
constraint that the flow be smooth.  Specifically it measures smoothness using laplacians, and seeks to minimise
the sum of the squares of the laplacians of u & v.  To combine the constraints together, we try to minimise:
integral over image of `(Ix.u+Iy.v+It)^2+lam(laplace^2 u + laplace^2 v)`

The laplacians are approximated by `laplace^2 u = (avg u - u)^2` for some average of the neighbourhood of u.
Assuming a gaussian distribution, this minimum occurs when its derivatives vanish, specifically when:

    Ix.(Ix.u+Iy.v+It)-lam(avg u - u) = 0
    Iy.(Ix.u+Iy.v+It)-lam(avg v - v) = 0

This leads to a linear system:

    A[u] = [-Ix.It]
     [v] = [-Ix.It]

Where A relates a pixel of u to itself by Ix^2+lam, to the same pixel of v by Ix.Iy, and to the pixels used
for avg u by the weights used to compute avg u scaled by -lam; similarly for v.  To compute the flow, we solve
this system.  Because A is sparse, we use an iterative technique, specifically conjugate gradient.  To estimate
Ix and Iy, we use the stencil [1 8 0 -8 -1]/12.  To estimate It, we simply subtract image 2 from image 1.
The average is computed by summing its four neighbours (and then dividing by four, except that the code below
seems to put the 1/4 into lam, and thus pixel is related to itself by Ix^2+4.lam).

Third, conjugate gradient is employed as follows.  Normally this algorithm would iterate until error is low
enough, instead we just use a fixed number of iterations (250 for last scale, 50 for rest).  The algorithm
with a preconditioner is as follows:

    x = initial guess of solution
    r = b-Ax
    z = Pr
    p = z
    repeat
      rsold = dot(r,z)
      alpha = rsold/dot(p,Ap)
      x += alpha.p
      r -= alpha.Ap
      z = Pr
      beta = dot(r,z)/rsold
      p = z+beta.p

Where P is the preconditioner (its inverse must be symmetric positive-definite).  The code below uses zero
as the initial guess.  The code below uses a matrix whose entry for a pixel of u to itself is Iy.Iy+4.lam, for u
to same pixel of v is Ix.Iy, and for v to itself is Ix.Ix+4.lam; except these entries are scaled by
1.0/((Ix.Ix+4.lam)(Iy.Iy+4lam)-(Ix.Iy)^2).  Finally, we do not store A in any form, nor do a sparse matrix vector
product.  Instead the Ap operation is done by computing a stencil of p, this is a six-point stencil - the pixel
itself, the pixel in the other component, and the four neightbours.

The algorithm in psuedo code is:

```
  for scale from coarsest to finest
    down sample images to current scale
    if first iteration
      flow = zeros (at coarsest scale)
      warped = second image
    else
      interpolate flow to current scale (from previous scale)
      warped = move second image back by flow
    delta flow = solve linear system for first image to warped image (see below)
    update flow by adding delta flow
single scale solution:
  Ix = stencil(i1, [1 8 0 -8 -1]/12 in x)
  Ix = stencil(i1, [1 8 0 -8 -1]/12 in x)
  It = i2-i1
  preconditioner(u, v) =
    det = ((Ix.Ix+4.lam)(Iy.Iy+4.lam)-(Ix.Iy)^2)
    u' = ((Iy.Iy+4.lam).u-Ix.Iy.v)/det
    v' = ((Ix.Ix+4.lam).v-Ix.Iy.u)/det
    (u', v')
  u & v = zeros
  ru = -Ix.It
  rv = -Iy.It
  (zu, zv) = preconditioner(ru, rv)
  (pu, pv) = (zu, zv)
  rsold = sum(ru.zu)+sum(rv.zv)
  repeat for num iterations
    (Apu, Apv) = apply six-point stencil above to pu, pv
    pTAp = sum(pu.Apu)+sum(pv.Apv)
    alpha = rsold/pTAp
    (xu, xv) += alpha.(pu, pv)
    (ru, rv) -= alpha.(ru, rv)
    (zu, zv) = preconditioner(pu, pv)
    rsnew = sum(pu.zu)+sum(pv.zv)
    beta = rsnew/rsold
    (pu, pv) = (zu, zv) + beta.(pu, pv)
```
