#include <cassert>
#include <cmath>
#include <omp.h>
#include <cstdio>
#include "common.h"
#include "Random.h"

void InitAssets( int numPaths, int numSteps, REALVEC* __restrict assetPtr, REAL spot, REAL vol, REAL maturity ) {
    assert(assetPtr);
    assert(numPaths%VLEN==0);
    __assume_aligned(assetPtr,64);
    const size_t numPathStrips = numPaths/VLEN;
#define asset(s,p) assetPtr[(s)*numPathStrips+(p)]
    
    REAL dt = maturity / REAL(numSteps);
    REAL volSqrtdt = vol*std::sqrt(dt);
    REAL fwdFactor = vol*vol*dt*-0.5;
    REALVEC* samplePtr = (REALVEC*) _mm_malloc(omp_get_max_threads()*numSteps*sizeof(REALVEC),64);

    // FIXME - figure out right way to parameterize seed    
    int seed = 12345;

#pragma omp parallel 
    {
        // Figure out on which strips current thread operates.
        int pfirst, plast;
        int id = GetStripInterval( pfirst, plast, numPathStrips );

        // Initialize random number generator on per-thread basis
        VSLStreamStatePtr stream;
        CheckVSL( vslNewStream(&stream, VSL_BRNG_MRG32K3A, seed) );
        CheckVSL( vslSkipAheadStream(stream, pfirst*numSteps*VLEN) );  

        // Generate random samples
        REALVEC* sample = samplePtr + id*numSteps;
        REAL mean = 0;
        REAL deviation = 1.0;

        // Note: using exp2(x * log2(e)) for exp(x), and multiply with log2(e) before
        REAL vsqrtdt_log2e = volSqrtdt * M_LOG2E;
        REAL fwdFactor_log2e = fwdFactor * M_LOG2E;
        for( int p=pfirst; p<plast; ++p ) {
            genGaussian(stream, numSteps*VLEN, (REAL*)sample, mean, deviation );
            asset(0,p)[:] = spot;
            for( int s=1; s<=numSteps; ++s ) 
                asset(s,p)[:] = asset(s-1,p)[:]*::exp2(fwdFactor_log2e + vsqrtdt_log2e * sample[s-1][:]);
        }

        // Delete random number generator
        vslDeleteStream(&stream);
    }
    _mm_free(samplePtr);
}
