#include "Timer.hpp"
#include "params.h"
#include <cmath>
#include <cstring>
#include <cstdlib>
#include <iomanip>
#include <iostream>
#include <cassert>
#include <omp.h>
#include "common.h"

const int progLoops = 1;

int main(int argc, char* argv[]) {
    int numPaths = 131072*4;
    int numSteps = 256;
    int numThreads = 0;
    if (argc > 1) 
        numPaths = strtol(argv[1],0,0);
    if (argc > 2)
        numSteps = strtol(argv[2],0,0);
    if (argc > 3) {
        numThreads = strtol(argv[3],0,0);
        omp_set_dynamic(0);
        omp_set_num_threads(numThreads);
    }
    if(numThreads == 0) {
        numThreads = omp_get_max_threads();
    }
    std::cout << "numPaths = " << numPaths << ", numSteps = " << numSteps << ", numThreads = " << numThreads << std::endl;

    /* ------------- option parameters      */
    REAL spot = 100.0;
    REAL strike = 110.0;
    REAL vol = 0.25;
    REAL maturity = 2.0;

    Timer tt;
    tt.start();

    REAL *cashFlowPut;
    cashFlowPut = (REAL*) _mm_malloc(sizeof(REAL)*numPaths, 64);

    REAL *assetPtr;
    assetPtr = (REAL*) _mm_malloc(sizeof(REAL*)*(numSteps+1)*numPaths, 64);
 #define asset(s,p) assetPtr[(s)*numPaths+(p)]
    assert( assetPtr );  

    double initAssetTime=0, backwardTime=0;
    REAL amerpayoffput = 0, finalputpayoff=0;

    WarmUp( numPaths, numSteps, (REALVEC*)assetPtr, (REALVEC*)cashFlowPut );
    tt.stop();
    std::cout << "Init time: " << tt.getTime() << " ms" << std::endl;  

    tt.start();
    for (int pLoop=0; pLoop<progLoops; pLoop++) {

        Timer ttt;
        ttt.start();
        InitAssets( numPaths, numSteps, (REALVEC*)assetPtr, spot, vol, maturity );
        ttt.stop();
        initAssetTime += ttt.getTime();

        ttt.start();
        GoBackInTime( numPaths, numSteps, (const REALVEC*)assetPtr, strike, (REALVEC*)cashFlowPut );
        ttt.stop();
        backwardTime +=  ttt.getTime();

        // Final results
        amerpayoffput = 0;
        finalputpayoff = 0;
#pragma omp parallel for simd reduction(+:amerpayoffput,finalputpayoff)
        for(int i=0; i < numPaths; i++) {
            amerpayoffput += cashFlowPut[i];
            finalputpayoff += std::max(strike - asset(numSteps,i), REAL(0));
        }
        amerpayoffput /= numPaths;
        finalputpayoff /= numPaths;
    }
    tt.stop();

    std::cout << "    init asset time = " << initAssetTime << std::endl;
    std::cout << "    backward time = " << backwardTime << std::endl;
    std::cout << "Computation time: " << tt.getTime() << " ms" << std::endl;
      
    std::cout <<std::setprecision(9) << "European Put Option Price: "<< finalputpayoff << std::endl;
    std::cout <<std::setprecision(9) << "American Put Option Price: "<< amerpayoffput << std::endl;

    // Divide milliseconds by 1000 to get seconds
    std::cout << "SELFTIMED " << tt.getTime() / 1000.0  << std::endl;
}

