#=
Copyright (c) 2015, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:
- Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice, 
  this list of conditions and the following disclaimer in the documentation 
  and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
THE POSSIBILITY OF SUCH DAMAGE.
=#

using DocOpt

const spot = 100.0
const strike = 110.0
const vol = 0.25
const maturity = 2.0

function init(numPaths::Int, numSteps::Int)
    # Pre-compute some values
    dt = maturity / numSteps
    volSqrtdt = vol * sqrt(dt)
    fwdFactor = vol^2 * dt * -0.5
    vsqrtdt_log2e = volSqrtdt * log2(exp(1.0))
    fwdFactor_log2e = fwdFactor * log2(exp(1.0))

    # Storage
    # Compute all per-path asset values for each time step
    # NOTE: hand-hoisting obscures the basic formula. 
    # Can the compiler do this optimization?
    asset = Array(Float64, numPaths, numSteps+1)
    fill!(asset, spot)
    for s=2:numSteps+1, p=1:numPaths
        asset[p,s] = asset[p,s-1] * exp2(fwdFactor_log2e + vsqrtdt_log2e * randn())
    end
    return asset
end

function model(strike, numPaths, numSteps, asset)
    avgPathFactor = 1.0/numPaths
    cashFlowPut = Array(Float64, 1, numPaths)
    for p = 1:numPaths
        cashFlowPut[1,p] = max(strike-asset[p,numSteps+1], 0.0) * avgPathFactor
    end

    # Now go back in time using regression
    regressionValPut = zeros(Float64, numPaths, 3)

    for s = numSteps:-1:2
        sum1 = 0.0; sum2=0.0; sum3=0.0; sum4=0.0
        sum0i = 0
        for p = 1:numPaths
            assetval = asset[p,s]

            if assetval <= strike
                sum0i += 1
                regressionValPut[p,1] = 1.0
                regressionValPut[p,2] = assetval
                regressionValPut[p,3] = assetval^2
                sum1 += assetval
                sum2 += assetval^2
                sum3 += assetval^3
                sum4 += assetval^4
            else
                regressionValPut[p,1] = 0
                regressionValPut[p,2] = 0
                regressionValPut[p,3] = 0
            end
        end

        # compute the regression
        ridgecorrect = 0.01
        sum0 = Float64(sum0i)
        sum0 += ridgecorrect
        sum1 += ridgecorrect
        sum2 += ridgecorrect
        sum3 += ridgecorrect
        sum4 += ridgecorrect
        sqrtest = Float64[sum0 sum1 sum2;
                          sum1 sum2 sum3;
                          sum2 sum3 sum4]
        invsqr = inv(sqrtest)
        betaPut = (cashFlowPut*regressionValPut)*invsqr
        for p=1:numPaths
            assetval = asset[p,s]
            if assetval<=strike
                regImpliedCashFlow = regressionValPut[p,1]*betaPut[1] +
                regressionValPut[p,2]*betaPut[2] +
                regressionValPut[p,3]*betaPut[3]
                payoff = strike-assetval
                if payoff > regImpliedCashFlow
                    cashFlowPut[1,p] = payoff
                end
            end
        end
    end

    amerpayoffput = 0.0
    finalputpayoff = convert(Float64,0.0)
    for i = 1:numPaths
        amerpayoffput  += cashFlowPut[1,i]
        finalputpayoff += max(strike-asset[i,numSteps+1], 0.0)
    end

    return amerpayoffput/numPaths, finalputpayoff/numPaths
end

function main()
    doc = """quant-devectorized.jl

Quantitative option pricing model (written in devectorized style).

Usage:
  quant-devectorized.jl -h | --help
  quant-devectorized.jl [--assets=<assets>] [--iterations=<iterations>]

Options:
  -h --help                  Show this screen.
  --assets=<assets>          Specify the number of assets to simulate [default: 524288].
  --iterations=<iterations>  Specify a number of iterations [default: 256].
"""
    arguments = docopt(doc)

    paths = parse(Int, arguments["--assets"])
    steps = parse(Int, arguments["--iterations"])

    srand(0)

    println("assets = ", paths)
    println("iterations = ", steps)

    asset = init(paths, steps)
    tic()
    model(strike, 1, 1, asset)
    compiletime = toq()
    println("SELFPRIMED ", compiletime)

    tic()
    amerpayoffput, finalputpayoff = model(strike, paths, steps, asset)
    selftimed = toq()
    println("European Put Option Price: ", finalputpayoff)
    println("American Put Option Price: ", amerpayoffput)
    println("SELFTIMED ", selftimed)

end

main()
