# A script to run ParallelAccelerator workloads in various
# configurations for testing or benchmarking purposes and print
# running times for each config that runs.

using DocOpt

envopts = [

           # round(Int, Base.CPU_CORES/2) should correspond to the
           # number of physical cores, not counting hyperthreading
           ("OMP_NUM_THREADS", 1, round(Int, Base.CPU_CORES/2)),

           # 0 = don't fuse any parfors
           # -1 = fuse maximum number of parfors
           ("PIRSetFuseLimit", 0, -1),

           # 0 = don't try to hoist allocations out of loops
           # 1 = try to hoist allocations out of loops
           ("PIRHoistAllocation", 0, 1),

           # 1 = vectorization off
           # 2 = vectorization on
           ("vectorizationlevel", 1, 2)
           ]

trials = 5

function println_flush(str)
    println(str)
    flush(STDOUT)
end

getenv(s::ASCIIString) = try x = parse(Int, ENV[s]) catch x = 0 end

function print_config(envopts)
    for opt in envopts
        println_flush(opt[1] * " = " * string(getenv(opt[1])))
    end
end

function reset(envopts)
    for opt in envopts
        ENV[opt[1]] = opt[2]
    end
end

function run_benchmarks(benchmarks_dir::UTF8String, workloads::Array{UTF8String})

    cd("$benchmarks_dir")

    # These should be on anyway from having run env-vars.sh, but it
    # doesn't hurt to make sure.
    ENV["KMP_AFFINITY"] = "scatter"
    ENV["OMP_NESTED"] = "TRUE"
    ENV["ReadFromEnv"] = "1"

    # Iterate over specified workloads, rather than all benchmarks.
    for program in workloads

        println_flush("BENCHMARK SHORT NAME: $program")

        reset(envopts)

        # First, do a baseline run with all opts off.
        println_flush("Starting configuration")
        print_config(envopts)

        cd("$program/benchmarking")

        println(STDERR, pwd())
        flush(STDERR)
        for i in 1:trials
            run(`./runjulia5-acc.sh`)
        end
        cd("../../")

        println_flush("Ending configuration")

        # Now iterate over opts, turning another one on with each
        # run.
        for opt in envopts
            println_flush("Starting configuration")

            o = opt[1]
            v = opt[3]
            ENV[o] = v
            print_config(envopts)

            cd("$program/benchmarking")
            for i in 1:trials
                run(`./runjulia5-acc.sh`)
            end
            cd("../../")

            println_flush("Ending configuration")
        end
    end
    ENV["ReadFromEnv"] = "0"
end

function main()
    doc = """configtest.jl

Usage:
  configtest.jl -h | --help
  configtest.jl [--benchmarks-dir=<path>] [--workloads=<workload-list>]

Options:
  -h --help                    Show this screen.
  --benchmarks-dir=<path>      Specify a path to the benchmarks directory; defaults to "..".
  --workloads=<workload-list>  Specify a list of workloads to run in the format 'workload1 workload2 ...'.
"""
    arguments = docopt(doc)

    if (arguments["--benchmarks-dir"] != nothing)
        benchmarks_dir = arguments["--benchmarks-dir"]
    else
        benchmarks_dir = ".."
    end

    if (arguments["--workloads"] != nothing)
        workloads = convert(Array{UTF8String}, split(arguments["--workloads"], " "))
    else
        workloads = convert(Array{UTF8String}, [key for key in keys(benchmarks)])
    end

    run_benchmarks(benchmarks_dir, workloads)

end

main()
