#!/usr/bin/env python

# munge-results.py: Process raw data from benchmarking runs and
# generate reports and plots.

import sys
import re
import copy
import os
import collections
import datetime
from optparse import OptionParser
from math import log10
from collections import OrderedDict
import matplotlib as mpl
mpl.use('Agg') # Tell matplotlib what backend to use.
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter
from matplotlib.colors import rgb2hex

def main():

    usage_string = """usage: munge-results.py totaltimes.log offloadtimes.log [--plot-verbose] [--opts-log=optimizations.log] [--filename-suffix=suffix] [--plot-font-size=n]
"""

    parser = OptionParser()
    parser.add_option("--plot-verbose", action="store_true", help="Tells munge-results to use long titles and subtitles on plots.", dest="plot_verbose")
    parser.add_option("--opts-log", help="Parse results from optimizations log.", dest="opts_log")
    parser.add_option("--filename-suffix", help="String to use as suffix for plot filenames.", dest="filename_suffix")
    parser.add_option("--plot-font-size", help="Number specifying font size to use on plot; defaults to 12.", dest="plot_font_size")

    (options, positional_args) = parser.parse_args()

    plot_verbose = options.plot_verbose
    opts_log = options.opts_log

    if options.filename_suffix != None:
        filename_suffix = options.filename_suffix
    else:
        filename_suffix = str(datetime.date.today())

    if options.plot_font_size != None:
        plot_font_size = int(options.plot_font_size)
    else:
        plot_font_size = 12

    # Always expect two positional args.
    if len(positional_args) != 2:
        print usage_string
        sys.exit(1)

    totaltimes_file = open(str(positional_args[0]), 'rU')
    offloadtimes_file = open(str(positional_args[1]), 'rU')

    results = parse_log_files(totaltimes_file, offloadtimes_file)
    average_running_times = get_average_times(results, "run")

    # "standard" mode: only munge data produced by benchmark.sh, and
    # produce a standard results plot.
    if opts_log is None:

        average_compile_times = get_average_times(results, "compile")
        validate_results(results, average_running_times)
        plot_results(results, average_running_times, plot_verbose, filename_suffix, plot_font_size)
        print_report(results, average_running_times, average_compile_times)

    # "opts" mode: additionally munge data produced by configtest.jl,
    # and produce an opts plot.
    else:
        opts_data = []

        opts_file = open(str(opts_log), 'rU')
        opts_results = parse_opts_file(opts_file)
        average_opts_running_times = get_average_opts_running_times(opts_results)
        plot_opts_results(results, average_running_times, opts_results, average_opts_running_times, plot_verbose, filename_suffix, plot_font_size)

def parse_log_files(totaltimes_file, offloadtimes_file):
    """Process log files and collect a record of timing data and benchmark metadata."""

    benchmark_name = "Name not specified"
    benchmark_description = "Description not specified"
    benchmark_short_name = "short-name-not-specified"
    totaltimes = []
    versions = []
    compiletimes = []
    offloadtimes = []
    runs_succeeded = [] # Array of bools indicating whether each run completed successfully
    compiles_succeeded = [] # Array of bools indicating whether each ParallelAccelerator compile completed successfully

    run_succeeded = False # Assume that the run failed until we see a SELFTIMED
    compile_succeeded = False # Assume that the compile failed (or that it just didn't happen) until we see one
    for line in totaltimes_file.readlines():
        if "SELFTIMED" in line:
            m = re.findall('\d+.\d+', line)
            num = float(m[0])
            totaltimes.append(num)
            run_succeeded = True
        elif "ParallelAccelerator.accelerate time:" in line:
            m = re.findall('\d+.\d+', line)
            num = float(m[0])
            compiletimes.append(num)
            compile_succeeded = True
        elif "SELFPRIMED" in line:
            m = re.findall('\d+.\d+', line)
            num = float(m[0])
            # Ignore this number if it's tiny.  These are cases where
            # we print a SELFPRIMED time but we aren't actually using
            # ParallelAccelerator.
            # N.B.: it would be nice to do this in some less ad-hoc way.
            if num > 0.5:
                compiletimes.append(num)
                compile_succeeded = True
        elif "Ending benchmark run" in line:
            runs_succeeded.append(run_succeeded)
            compiles_succeeded.append(compile_succeeded)

            # Reset for next run
            run_succeeded = False
            compile_succeeded = False
        else:
            pass

    offload_total_for_run = 0
    for line in offloadtimes_file.readlines():
        if "BENCHMARK NAME:" in line:
            line = re.sub('BENCHMARK NAME: ', '', line)
            benchmark_name = line.rstrip()
        elif "BENCHMARK DESCRIPTION:" in line:
            line = re.sub('BENCHMARK DESCRIPTION: ', '', line)
            benchmark_description = line.rstrip()
        elif "BENCHMARK SHORT NAME:" in line:
            line = re.sub('BENCHMARK SHORT NAME: ', '', line)
            benchmark_short_name = line.rstrip()
        elif "Version:" in line:
            line = re.sub('Version: ', '', line)
            line = re.sub(':', '', line)
            versions.append(line.rstrip())
        elif "Starting benchmark run" in line:
            offload_total_for_run = 0
        elif "offload kernel time = " in line:
            m = re.findall('\d+.\d+', line)
            num = float(m[0])
            offload_total_for_run = offload_total_for_run + num
        elif "Segmentation fault" in line:
            # The run failed
            print "Warning: a run segfaulted"
        elif "License checkout failed" in line:
            print "Warning: license checkout failed"
        elif "Target architecture mismatch" in line:
            print "Warning: target architecture mismatch"
        elif "Ending benchmark run" in line:
            offloadtimes.append(offload_total_for_run)

            # Reset for next run
            offload_total_for_run = 0
        else:
            pass

    # If a run failed, there won't be a total time for it in totaltimes.
    # So we need to add a marker to totaltimes saying that the run failed.
    for i in range(len(runs_succeeded)):
        if (runs_succeeded[i]):
            # this one succeeded, so we don't need to do anything
            pass
        else:
            # otherwise, indicate that the run failed
            totaltimes.insert(i, "(run failed)")

    # Similarly, we need to add a marker to compiletimes to say that
    # there's no compile data.

    # N.B.: Ideally, we'd distinguish between compiles that didn't
    # happen (for non-@acc versions) and compiles that actually
    # failed.  For now, they are all just 'None'.
    for i in range(len(compiles_succeeded)):
        if (compiles_succeeded[i]):
            # this one succeeded, so we don't need to do anything
            pass
        else:
            # otherwise, indicate that there was no compile
            compiletimes.insert(i, None)

    # now return a record of the data structures we built
    return {
        "benchmark_name": benchmark_name,
        "benchmark_short_name": benchmark_short_name,
        "benchmark_description": benchmark_description,
        "totaltimes": totaltimes,
        "versions": versions,
        "offloadtimes": offloadtimes,
        "compiletimes": compiletimes,
        "runs_succeeded": runs_succeeded,
        "compiles_succeeded": compiles_succeeded,
    }

def parse_opts_file(opts_file):
    """Process an opts log file and collect a record of timing data."""

    results = {}
    results['benchmark_short_name'] = None
    results['baseline'] = None
    results['parallel'] = None
    results['fused'] = None
    results['hoisted'] = None
    results['vectorized'] = None

    result = {}
    result['PIRNumThreadsMode'] = None
    result['OMP_NUM_THREADS'] = None
    result['PIRSetFuseLimit'] = None
    result['PIRHoistAllocation'] = None
    result['vectorizationlevel'] = None

    configtimes = []

    for line in opts_file.readlines():
        if "BENCHMARK SHORT NAME:" in line:
            line = re.sub('BENCHMARK SHORT NAME: ', '', line)
            results['benchmark_short_name'] = line.rstrip()
        elif "PIRNumThreadsMode =" in line:
            m = re.findall('\d+', line)
            num = int(m[0])
            result['PIRNumThreadsMode'] = num
        elif "OMP_NUM_THREADS =" in line:
            m = re.findall('\d+', line)
            num = int(m[0])
            result['OMP_NUM_THREADS'] = num
        elif "PIRSetFuseLimit =" in line:
            m = re.findall('-?\d+', line)
            num = int(m[0])
            result['PIRSetFuseLimit'] = num
        elif "PIRHoistAllocation =" in line:
            m = re.findall('-?\d+', line)
            num = int(m[0])
            result['PIRHoistAllocation'] = num
        elif "vectorizationlevel =" in line:
            m = re.findall('-?\d+', line)
            num = int(m[0])
            result['vectorizationlevel'] = num
        elif "SELFTIMED" in line:
            m = re.findall('\d+.\d+', line)
            num = float(m[0])
            configtimes.append(num)
        elif "Ending configuration" in line:

            # At this point, 'result' should be filled in, so we just
            # need to find out what configuration this is and then add
            # an entry to `results` for it.
            if is_baseline(result):
                results["baseline"] = configtimes
            elif is_parallel(result):
                results["parallel"] = configtimes
            elif is_fused(result):
                results["fused"] = configtimes
            elif is_hoisted(result):
                results["hoisted"] = configtimes
            elif is_vectorized(result):
                results["vectorized"] = configtimes
            else:
                print "warning: unexpected configuration: ", result

            # Reset for next configuration
            result['PIRNumThreadsMode'] = None
            result['OMP_NUM_THREADS'] = None
            result['PIRSetFuseLimit'] = None
            result['PIRHoistAllocation'] = None
            result['vectorizationlevel'] = None
            configtimes = []

        else:
            pass

    return results

def is_baseline(result):
    return (result['OMP_NUM_THREADS'] == 1 and
            result['PIRSetFuseLimit'] == 0 and
            result['PIRHoistAllocation'] == 0 and
            result['vectorizationlevel'] == 1)

def is_parallel(result):
    return (result['OMP_NUM_THREADS'] == 36 and
            result['PIRSetFuseLimit'] == 0 and
            result['PIRHoistAllocation'] == 0 and
            result['vectorizationlevel'] == 1)

def is_fused(result):
    return (result['OMP_NUM_THREADS'] == 36 and
            result['PIRSetFuseLimit'] == -1 and
            result['PIRHoistAllocation'] == 0 and
            result['vectorizationlevel'] == 1)

def is_hoisted(result):
    return (result['OMP_NUM_THREADS'] == 36 and
            result['PIRSetFuseLimit'] == -1 and
            result['PIRHoistAllocation'] == 1 and
            result['vectorizationlevel'] == 1)

def is_vectorized(result):
    return (result['OMP_NUM_THREADS'] == 36 and
            result['PIRSetFuseLimit'] == -1 and
            result['PIRHoistAllocation'] == 1 and
            result['vectorizationlevel'] == 2)

def get_average_times(results, trial_type):
    """Compute average running or compile times for each version of a benchmark."""

    if (trial_type == "compile"):
        trials_succeeded = results["compiles_succeeded"]
        trial_data = results["compiletimes"]
    elif (trial_type == "run"):
        trials_succeeded = results["runs_succeeded"]
        trial_data = results["totaltimes"]
    else:
        assert False, "unexpected trial_type"

    benchmark_short_name = results["benchmark_short_name"]
    versions = results["versions"]

    average_times = []
    trials_per_version = len(trials_succeeded) / len(versions)

    for i in range(len(trial_data)):
        trial_number = (i % trials_per_version) + 1

        # Which version are we on?  Ugh.  This kind of nonsense
        # wouldn't be necessary if we were using nicer data
        # structures...
        version = results['versions'][i / (len(versions) + 1)]

        # is this the last trial for this version?
        if (trial_number == trials_per_version):
            first_trial_index = i - trials_per_version + 1

            trial_data_for_this_version = trial_data[first_trial_index:i+1]

            # Assume all trials succeeded...
            successful_trials = trials_per_version

            # ...then check for failed trials:

            # Failed compile.
            if (None in trial_data_for_this_version and trial_type == "compile"):
                trial_data_successful = filter(lambda x: x != None, trial_data_for_this_version)
                successful_trials = len(trial_data_successful)

            # Failed run.
            elif ("(run failed)" in trial_data_for_this_version and trial_type == "run"):
                trial_data_successful = filter(lambda x: x != '(run failed)', trial_data_for_this_version)
                successful_trials = len(trial_data_successful)

            # If there were no successful trials and we're measuring
            # compile time, then it's probably just a non-@acc
            # version (for which there is no compile-time data), so
            # just skip it.
            if (successful_trials == 0 and trial_type == "compile"):
                average_time = None

            # If there were no successful trials and we're measuring
            # running time, use 10s and print a warning.
            elif (successful_trials == 0 and trial_type == "run"):
                print "Warning: there were no successful runs of version '%s' of '%s' -- using 5s instead." % (version, benchmark_short_name)
                average_time = 5

            # All trials succeeded, so compute an average dropping
            # first and last trial.
            elif (successful_trials == trials_per_version):
                average_time = sum(trial_data[first_trial_index+1:i]) / float(trials_per_version-2)
            # Some but not all trials succeeded, so include
            # everything.
            else:
                average_time = sum(trial_data_successful) / float(successful_trials)

            average_times.append(average_time)

    return average_times

def get_average_opts_running_times(results):

    benchmark_short_name = results['benchmark_short_name']
    average_running_times = {}

    for version, times in results.iteritems():
        if version == 'benchmark_short_name':
            continue

        if times == None:
            print "Warning: no results collected for '%s' version." % version
            successful_runs = 0
        else:
            successful_runs = len(times)

        # N.B.: It would be nice if this weren't hard-coded.
        # If all runs succeeded, compute an average dropping first and last run.
        if (successful_runs == 5):
            average_running_time = sum(times[1:successful_runs-1]) / float(successful_runs-2)
        # Otherwise, include everything -- unless EVERY run
        # failed, in which case return 0 and print a warning.
        elif (successful_runs == 0):
            print "Warning: there were no successful runs of version '%s' of '%s' -- using 0.00001 instead." % (version, benchmark_short_name)
            average_running_time = 0.00001
        else:
            average_running_time = sum(times) / float(successful_runs)

        average_running_times[version] = average_running_time

    return average_running_times

def validate_results(results, average_running_times):
    """Check that certain expected properties of the timing data hold."""
    runs_succeeded = results["runs_succeeded"]
    compiles_succeeded = results["compiles_succeeded"]
    versions = results["versions"]
    compiletimes = results["compiletimes"]

    # The number of runs should be an integer multiple of the number of
    # versions, since we did N runs for each version.
    assert len(runs_succeeded) >= len(versions), """
    Something's wrong: you seem to have fewer runs than versions, and
    there should be at least one run per version.  Check to make sure
    you're logging elapsed time for each run.
    """

    assert len(versions) == len(average_running_times), """
    Something's wrong: you don't have the same number of average
    running times as you have versions.
    """

    assert len(runs_succeeded) == len(compiles_succeeded), """
    Something's wrong: you have a different number of attempted runs
    and attempted compiles.
    """

def plot_setup(x_data, y, benchmark_short_name, figure_title, filename_suffix, plot_font_size,
               mode="standard", extra_labels=None, label_fontsize=18):

    def log_10_product(x, pos):
        """The two args are the value and tick position.
        Label ticks with the product of the exponentiation"""
        if x <= 0.1:
            # Prevent ticks less than 0 (0.1, 0.01, ...) from just being rounded to 0
            return x
        else:
            return '%1i' % (x)

    # Set some matplotlibrc params to make the plot prettier.
    mpl.rcParams['font.size'] = plot_font_size
    mpl.rcParams['axes.titlesize'] = plot_font_size
    mpl.rcParams['axes.labelsize'] = plot_font_size
    mpl.rcParams['figure.figsize'] = [11, 8]

    f = plt.figure()
    x = xrange(len(x_data))

    # [left, bottom, width, height] where all quantities are in fractions of figure width and height. 
    ax = f.add_axes([0.2, 0.4, 0.75, 0.5])
    ax.spines['top'].set_visible(False)

    ax.set_yscale('log')
    ylabel = "Speedup"
    formatter = FuncFormatter(log_10_product)
    ax.yaxis.set_major_formatter(formatter)

    plt.axhline(1.0, color="#555555", linestyle="dashed") # show a baseline line
    separate_legend = True # show legend as a separate figure

    # Various mode-specific choices.
    if (mode == "opts"):
        # Colors used for opts mode.  This is a lot simpler because in
        # opts mode we always have the same six bars on the plot.
        # Colors selected with the help of colorbrewer2.org
        bar_colors = ['#ccebc5', '#a8ddb5', '#7bccc4', '#4eb3d3', '#2b8cbe', '#08589e']
        filename_end = "-opts.pdf"


    elif (mode == "standard"):
        # Colors for standard mode have to be chosen on a case-by-case
        # basis from color_dict, depending which versions are in the
        # data at hand.

        # Mapping of version labels to colors.
        color_dict = OrderedDict([
            ("Python",                           '#f89bd4'),
            ("Matlab (1 thread)",                '#fc8d62'),
            ("Matlab",                           '#b84920'),
            ("Julia (1 thread)",                 '#ccebc5'),
            ("Julia (explicit loops, 1 thread)", '#88a781'),
            ("@acc (1 thread)",                  '#4c9cdf'),
            ("@acc (36 threads)",                '#08589e'),
            ("C (simple serial, 1 thread)",      '#a784e6'),
            ("C/C++ (optimized, 36 threads)",    '#6340a2')])

        # Choose colors from color_dict based on x_data.  For some
        # reason, we have to put this inside an eval, or matplotlib
        # produces an error. D: D:
        bar_colors = eval("[color_dict[x] for x in x_data]")
        filename_end = ".pdf"
    else:
        assert False, "unknown mode"

    rects = ax.bar(x, y, align='center', linewidth=0, color=bar_colors)
    ax.set_xticks(x)
    f.autofmt_xdate()

    # Don't show tick labels on x-axis
    ax.set_xticklabels([])

    if separate_legend:
        pass
    else:
        ax.legend(rects, x_data,
              # Controls legend placement
              'upper right',
              # Controls font size in legend
              prop={'size':16})

    # Remove ticks from x-axis
    for tic in ax.xaxis.get_major_ticks():
        tic.tick1On = tic.tick2On = False

    plt.title(figure_title)
    plt.ylabel(ylabel)

    if mode == "opts":
        # Tweak scaling for nengo.
        ylim_bottom, ylim_top = ax.get_ylim()
        if ylim_top == 1.0:
            ax.set_ylim(ylim_bottom, 2.0)

        autolabel_opts(rects, ax, y, plt, label_fontsize, extra_labels)
    else:
        autolabel(rects, ax, y, plt, label_fontsize, extra_labels)

    file_path = "/tmp/results/" + benchmark_short_name + "/"

    if not os.path.exists(file_path):
        os.makedirs(file_path)

    filename = file_path + benchmark_short_name + "-" + filename_suffix + filename_end
    plt.savefig(filename)

def autolabel(rects, ax, data, plt, label_fontsize, extra_labels):
    """Put values from `data` as a number at the top of each column."""
    # Get y-axis top limit to calculate label position from.
    (y_limit_bottom, y_limit_top) = ax.get_ylim()
    y_height = y_limit_top - y_limit_bottom

    for i,rect in enumerate(rects):
        label_color = "black"

        color_rgba = rect._facecolor # RGBA representation
        if all(i >= 0.26 for i in color_rgba):
            # light enough color; use dark label
            inner_label_color = "black"
        else:
            # dark color; use white label
            inner_label_color = "white"

        height = rect.get_height()

        # Fraction of height taken up by this rectangle
        f_height = (height / y_height)

        # We'll just use this until I figure out what else to do. :(
        label_position = height * 1.06

        # Add speedup labels
        plt.text(rect.get_x() + rect.get_width()/2.,
                 label_position,
                 '%.1fx' % data[i],
                 ha='center',
                 va='bottom',
                 fontsize=label_fontsize,
                 color=label_color)

        # Add running time labels
        plt.text(rect.get_x() + rect.get_width()/2.,
                 y_limit_bottom * 1.001,
                 '(%.1f s)' % extra_labels[i],
                 ha='center',
                 va='bottom',
                 fontsize=label_fontsize-4,
                 color=inner_label_color)

def autolabel_opts(rects, ax, data, plt, label_fontsize, extra_labels):
    """Put values from `data` as a number at the top of each column, and
    values from `extra_labels` at the bottom."""

    # Get y-axis top limit to calculate label position from.
    (y_limit_bottom, y_limit_top) = ax.get_ylim()
    y_height = y_limit_top - y_limit_bottom

    for i,rect in enumerate(rects):
        label_color = "black"

        color_rgba = rect._facecolor # RGBA representation
        if all(i >= 0.1 for i in color_rgba):
            # light enough color; use dark label
            inner_label_color = "black"
        else:
            # dark color; use white label
            inner_label_color = "white"

        height = rect.get_height()

        # Fraction of height taken up by this rectangle
        f_height = (height / y_height)

        label_position = height * 1.06

        # Add speedup labels; don't put a label on the first column
        if i != 0:
            plt.text(rect.get_x() + rect.get_width()/2.,
                     label_position,
                     '%.1fx' % data[i],
                     ha='center',
                     va='bottom',
                     fontsize=label_fontsize,
                     color=label_color)

        # Add running time labels
        plt.text(rect.get_x() + rect.get_width()/2.,
                 y_limit_bottom * 1.001,
                 '(%.1f s)' % extra_labels[i],
                 ha='center',
                 va='bottom',
                 fontsize=label_fontsize-4,
                 color=inner_label_color)

def plot_results(results, average_running_times, plot_verbose,
                 filename_suffix, plot_font_size):
    """Plot timing data using matplotlib and save the plot to a file."""

    benchmark_name = results["benchmark_name"]
    benchmark_description = results["benchmark_description"]
    benchmark_short_name = results["benchmark_short_name"]
    versions = results["versions"]

    # If we have more than 8 versions on the plot, make number labels
    # smaller to fit.
    if len(versions) >= 8:
        label_fontsize = plot_font_size
    else:
        label_fontsize = 18

    # Order we want to show the final versions in.
    sorted_versions = [
        "Python",
        "Matlab (1 thread)",
        "Matlab",
        "Julia (1 thread)",
        "Julia (explicit loops, 1 thread)",
        "@acc (1 thread)",
        "@acc (36 threads)",
        "C (simple serial, 1 thread)",
        "C/C++ (optimized, 36 threads)"
    ]

    # Normalize names into above set of names.
    versions = [re.sub('Julia 0.5 \(1 thread\)', 'Julia (1 thread)', version) for version in versions]
    versions = [re.sub('Julia 0.5 \(devec\)', 'Julia (explicit loops, 1 thread)', version) for version in versions]
    versions = [re.sub('Julia 0.5', 'Julia (1 thread)', version) for version in versions]
    versions = [re.sub('Julia 0.4', 'Julia (1 thread)', version) for version in versions]
    versions = [re.sub('Prospect ', '', version) for version in versions]
    versions = [re.sub('single-threaded', '1 thread', version) for version in versions]
    versions = [re.sub('C\+\+ \(optimized, 36 threads\)', 'C/C++ (optimized, 36 threads)', version) for version in versions] 
    versions = [re.sub('C \(optimized SSE', 'C/C++ (optimized', version) for version in versions] 
    versions = [re.sub('^C\+\+$', 'C/C++ (optimized, 36 threads)', version) for version in versions] 

    zipped_lists = zip(versions, average_running_times)

    julia_only = filter(lambda x : x[0] == "Julia (1 thread)", zipped_lists)
    julia_average_running_time = julia_only[0][1]

    def remove_unwanted_versions(x):
        """Remove versions we don't want to show on plots."""

        # stuff we're leaving out of the paper
        return "@threads" not in x[0] and "18 threads" not in x[0] and \
        "simple parallel" not in x[0] and "AVX" not in x[0] and \
        "@acc threads" not in x[0]

    filtered_zipped_lists = filter(remove_unwanted_versions, zipped_lists)

    # Sort according to order we want to show them in.
    filtered_zipped_lists = sorted(filtered_zipped_lists, key = lambda x: sorted_versions.index(x[0]))

    unzipped_lists = zip(*filtered_zipped_lists)
    abridged_versions = unzipped_lists[0]
    abridged_average_running_times = unzipped_lists[1]

    baseline = julia_average_running_time
    speedups = map(lambda time: baseline / time, abridged_average_running_times)

    # Treat @acc versions like regular versions
    abridged_versions = [re.sub(' @acc', '', version) for version in abridged_versions]

    # If we're doing verbose plotting, use the long name and
    # description; otherwise, don't.
    if plot_verbose:
        figure_title = benchmark_name + "\n" + benchmark_description
    else:
        figure_title = benchmark_short_name

    plot_setup(abridged_versions, speedups, benchmark_short_name,
               figure_title, filename_suffix, plot_font_size,
               mode="standard", label_fontsize=label_fontsize, extra_labels=abridged_average_running_times)

def plot_opts_results(results, average_running_times, opts_results,
                      average_opts_running_times, plot_verbose, filename_suffix,
                      plot_font_size):
    """Plot opts-mode timing data using matplotlib and save the plot to a file."""

    benchmark_name = results["benchmark_name"]
    benchmark_description = results["benchmark_description"]
    benchmark_short_name = results["benchmark_short_name"]

    # Grab just the Julia results, so we can add them to this figure.
    versions = results["versions"]
    zipped_lists = zip(versions, average_running_times)
    julia_only = filter(lambda x : x[0] == "Julia 0.5" or x[0] == "Julia 0.5 (1 thread)" or x[0] == "Julia 0.4", zipped_lists)

    julia_average_running_time = julia_only[0][1]

    acc_only = filter(lambda x : "@acc (36 threads)" in x[0], zipped_lists)

    acc_average_running_time = acc_only[0][1]

    figure_title = benchmark_short_name

    # N.B.: The rightmost bar in this figure will be from the regular
    # benchmark run, not from the optimizations run.  This is done so
    # that the two figures will agree with each other.
    running_times = [julia_average_running_time,
                     average_opts_running_times["baseline"],
                     average_opts_running_times["parallel"],
                     average_opts_running_times["fused"],
                     average_opts_running_times["hoisted"],
                     acc_average_running_time]

    baseline = julia_average_running_time
    speedups = map(lambda time: baseline / time, running_times)

    versions = ["Julia (1 thread)",
                "@acc (1 thread, no opts)",
                "+Parallelization",
                "+Fusion",
                "+Hoisting",
                "+Vectorization"]

    # If we're doing verbose plotting, use the long name and
    # description; otherwise, don't.
    if plot_verbose:
        figure_title = benchmark_name + "\n" + benchmark_description
    else:
        figure_title = benchmark_short_name

    plot_setup(versions, speedups, benchmark_short_name, figure_title, filename_suffix,
               plot_font_size, mode="opts", extra_labels=running_times)

def print_report(results, average_running_times, average_compile_times):
    """Analyze timing data and print a report to stdout."""

    print """
    Note: if all runs succeeded, we drop the first and last runs when
    computing the average.  If some runs failed, we include all successful
    runs in the average.
    """

    benchmark_short_name = results["benchmark_short_name"]
    totaltimes = results["totaltimes"]
    versions = results["versions"]
    compiletimes = results["compiletimes"]
    offloadtimes = results["offloadtimes"]
    runs_succeeded = results["runs_succeeded"]

    version_count = 0
    runs_per_version = len(runs_succeeded) / len(versions)

    print "Benchmarking report for", benchmark_short_name

    for i in range(len(totaltimes)):
        # is this a new version?
        if (i % runs_per_version == 0):
            print "=================================================\n"
            print versions[version_count]

        run_number = (i % runs_per_version) + 1

        print "Run", run_number
        print "   running time:              ", totaltimes[i]

        if versions[version_count] == "@acc (36 threads)" or versions[version_count] == "@acc (1 thread)":
            # This shouldn't be 'None' unless the compile failed.
            print "   compile time (@acc):       ", compiletimes[i]

        # if we actually got a number, show more detailed info
        if totaltimes[i] != "(run failed)":
            if offloadtimes[i] == 0:
                print "   (no offloaded computation)"
            else:
                print "   offload computation time:", offloadtimes[i]
                print "   communication time:      ", totaltimes[i] - offloadtimes[i]

        # is this the last run for this version?
        if (run_number == runs_per_version):
            average_running_time = average_running_times[version_count]
            print "\nAverage running time over", runs_per_version, "runs:", average_running_time

            average_compile_time = average_compile_times[version_count]
            if average_compile_time != None:
                print "Average compile time over", runs_per_version, "runs:", average_compile_time

            # Increment the version count before going to the next version.
            version_count = version_count + 1

    print "=================================================\n"

if __name__ == "__main__":
    main()
